""" Utilities """
from petlib.ec import EcGroup



# ==================================================
# El-Gamal encryption scheme
# ==================================================
def elgamal_keygen(params):
   """ generate an El Gamal key pair """
   (G, o, g, h) = params
   d = o.random()
   gamma = d * g
   return (d, gamma)

def elgamal_enc(params, gamma, m, h):
    """ encrypts the values of a message (h^m) """
    (G, o, g, _) = params
    k = o.random()
    a = k*g
    b = k*gamma + m*h
    return (a, b, k)

def elgamal_dec(d, c):
    """ decrypts the message (h^m) """
    (a, b) = c
    return b - d*a


# ==================================================
# other
# ==================================================
def ec_sum(list):
	""" sum EC points list """
	ret = list[0]
	for i in range(1,len(list)):
		ret = ret + list[i]
	return ret






""" amac credential scheme based on mac ggm """
from petlib.ec import EcGroup
from amac.mac_ggm import setup as setup_ggm
from amac.mac_ggm import keygen as ggm_keygen
from amac.mac_ggm import mac as ggm_mac
from amac.utils import *
from amac.proofs import *


def setup():
	""" generate all public parameters """
	return setup_ggm()


def cred_keygen(params, q):
	""" credential keygen """
	assert q > 0
	(G, o, g, h) = params
	(x, X) = ggm_keygen(params, q)
	x0_tilde = o.random()
	Cx0 = x[0]*g + x0_tilde*h
	iparams = (Cx0, X)
	sk = (x, x0_tilde)
	return (iparams, sk)
	

def prepare_blind_sign(params, gamma, private_m):
	""" prepare issuance of credentials """
	assert gamma and len(private_m) > 0
	(G, o, g, h) = params
	enc = [elgamal_enc(params, gamma, m, g) for m in private_m]
	(a, b, k) = zip(*enc)
	c = list(zip(a, b))
	pi_prepare_issue = make_pi_prepare_issue(params, gamma, c, k, private_m)
	return (c, pi_prepare_issue)


def blind_issue(params, sk, iparams=None, gamma=None, c=[], pi_prepare_issue=None, public_m=[]):
	""" issue credentials """
	assert len(c)+len(public_m) > 0
	(G, o, g, h) = params
	(x, x0_tilde) = sk
	assert x0_tilde and len(x) > len(c)+len(public_m)

	## mixed public / private attributes
	if len(c): 
		assert iparams and gamma and pi_prepare_issue
		assert verify_pi_prepare_issue(params, gamma, c, pi_prepare_issue)
		(Cx0, X) = iparams
		(enc_a, enc_b) = zip(*c) 
		# blinding factors
		b = o.random()
		u = b*g
		bx0_tilde = b*x0_tilde % o
		bx = [b*xi % o for xi in x]
		bCx0 = b*Cx0
		bX = [b*Xi for Xi in X]
		bsk = (bx, bx0_tilde)
		biparams = (bCx0, bX)
		# create credential over encrypted attributes	
		r = o.random() 
		t1 = [mi*g for mi in public_m]
		t2 = r*g + ec_sum([bxi*enc_ai for (bxi,enc_ai) in zip(bx[1:],enc_a)])
		t3 = r*gamma + bx[0]*g + ec_sum([bxi*enc_bi for (bxi,enc_bi) in zip(bx[1:],list(enc_b)+t1)])
		u_prime = (t2, t3)
		pi_issue = make_pi_issue(params, sk, iparams=iparams, gamma=gamma, c=c, b=b, bsk=bsk, r=r, public_m=public_m)
		return (u, u_prime, pi_issue, biparams)

	## only public attributes
	else: 
		(u, u_prime) = ggm_mac(params, x, public_m)
		pi_issue = make_pi_issue(params, sk, u=u, public_m=public_m)
		return (u, u_prime, pi_issue)
		

def blind_obtain(params, iparams, u, u_prime, pi_issue, biparams=None, gamma=None, c=[], public_m=[]):
	""" verify credentials issuance """
	assert len(c)+len(public_m) > 0
	# verify proof of issuance
	return verify_pi_issue(params, iparams, u, u_prime, pi_issue, biparams=biparams, gamma=gamma, 
		ciphertext=c, public_m=public_m)


def unblind(d, u_prime):
	""" unblind credentials on private attributes """
	assert d and u_prime
	return elgamal_dec(d, u_prime)


def show(params, iparams, cred, private_m=[], public_m=[]):
	""" show credentials """
	attributes = private_m + public_m
	assert cred and len(attributes) > 0
	(G, o, g, h) = params
	(u0, u0_prime) = cred
	assert u0 and u0_prime
	# randomize credentials
	a = o.random()
	(u, u_prime) = (a*u0, a*u0_prime)
	# form commitments
	r = o.random()
	z = [o.random() for _ in attributes]
	Cm = [mi*u + zi*h for (mi,zi) in zip(attributes,z)]
	Cu_prime = u_prime + r*g
	sigma = (u, Cm, Cu_prime)
	pi_show = make_pi_show(params, iparams, attributes, r, z, u_prime, sigma)
	return (sigma, pi_show)


def show_verify(params, sk, iparams, sigma, pi_show):
	""" verify credentials """
	assert iparams and sigma and pi_show
	(x, _) = sk
	(u, Cm, Cu_prime) = sigma
	assert u and len(Cm) > 0 and Cu_prime
	V = x[0]*u + ec_sum([xi*Cmi for (xi,Cmi) in zip(x[1:],Cm)]) - Cu_prime
	return verify_pi_show(params, iparams, sigma, pi_show, V)









""" aMAC zero-knowledge proofs. """
from petlib.bn import Bn
from petlib.ec import EcGroup
from hashlib import sha256
from binascii import hexlify
from amac.utils import ec_sum


def to_challenge(elements):
    """ generates a Bn challenge by hashing a number of EC points """
    Cstring = b",".join([hexlify(x.export()) for x in elements])
    Chash = sha256(Cstring).digest()
    return Bn.from_binary(Chash)


def make_pi_prepare_issue(params, gamma, ciphertext, k, private_m):
	""" make prepare issuance proof """
	(G, o, g, h) = params
	assert len(ciphertext) == len(k) and len(ciphertext) == len(private_m)
	wk = [o.random() for _ in k]
	wm = [o.random() for _ in private_m]
	# compute the witnesses commitments
	Aw = [wki*g for wki in wk]
	Bw = [wk[i]*gamma + wm[i]*g for i in range(len(private_m))]
	# create the challenge
	(a, b) = zip(*ciphertext)
	c = to_challenge([g, h]+list(a)+list(b)+Aw+Bw)
	# create responses
	rk = [(wk[i] - c*k[i]) % o for i in range(len(wk))]
	rm = [(wm[i] - c*private_m[i]) % o for i in range(len(wm))]
	return (c, rk, rm)


def verify_pi_prepare_issue(params, gamma, ciphertext, proof):
	""" verify prepare issuance proof """
	(G, o, g, h) = params
	(c, rk, rm) = proof
	assert c and len(ciphertext) == len(rk) and len(ciphertext) == len(rm)
	(a, b) = zip(*ciphertext)
	# re-compute witnesses commitments
	Aw = [c*a[i] + rk[i]*g for i in range(len(rk))]
	Bw = [c*b[i] + rk[i]*gamma + rm[i]*g for i in range(len(ciphertext))]
	# verify challenge
	return c == to_challenge([g, h]+list(a)+list(b)+Aw+Bw)


def make_pi_issue(params, sk, iparams=None, gamma=None, c=[], b=None, bsk= None, r=None, public_m=[], u=None):
	""" make issuance proof """
	assert len(c)+len(public_m) > 0
	(G, o, g, h) = params
	(x, x0_tilde) = sk
	assert x0_tilde and len(x) > len(c)+len(public_m)

	## mixed public / private attributes
	if len(c): 
		assert iparams and gamma and b and r
		(bx, bx0_tilde) = bsk
		assert bx0_tilde and len(bx) == len(x)
		(Cx0, X) = iparams
		assert Cx0 and len(x) == len(X)+1
		(enc_a, enc_b) = zip(*c) 
		# create the witnesses
		wx = [o.random() for _ in x]
		wx0_tilde = o.random()
		wb = o.random()
		wbx = [o.random() for _ in x]
		wbx0_tilde = o.random()
		wr = o.random()
		# compute the witnesses commitments (prove knowledge of MAC's secrets)
		Aw = wx[0]*g + wx0_tilde*h
		Bw = wb*g
		Cw = wbx[0]*g + wbx0_tilde*h
		Dw = wb*Cx0
		# compute the witnesses commitments (prove of correctness of Xi's)
		Ew = [wxi*h for wxi in wx[1:]]
		Fw = [wbxi*h for wbxi in wbx[1:]]
		Gw = [wb*Xi for Xi in X]
		# compute the witnesses commitments (prove of correctness of credentials)
		t1 = [mi*g for mi in public_m]
		Hw = wr*g + ec_sum([wbxi*enc_ai for wbxi,enc_ai in zip(wbx[1:],enc_a)])
		Iw = wr*gamma + wbx[0]*g + ec_sum([wbxi*enc_bi for wbxi,enc_bi in zip(wbx[1:],list(enc_b)+t1)])
		# create the challenge
		c = to_challenge([g, h, Aw, Bw, Cw, Dw, Hw, Iw]+Ew+Fw+Gw)
		# create responses
		rx = [(wxi - c*xi) % o for (wxi,xi) in zip(wx,x)]
		rx0_tilde = (wx0_tilde - c*x0_tilde) % o
		rb = (wb - c*b) % o
		rbx = [(wbxi - c*bxi) % o for (wbxi,bxi) in zip(wbx,bx)]
		rbx0_tilde = (wbx0_tilde - c*bx0_tilde) % o
		rr = (wr - c*r) % o
		return (c, rx, rx0_tilde, rb, rbx, rbx0_tilde, rr)

	## only public attributes
	else: 
		assert u
		# create the witnesses
		wx = [o.random() for _ in x]
		wx0_tilde = o.random()
		# compute the witnesses commitments
		Aw = wx[0]*g + wx0_tilde*h
		Bw = [wxi*h for wxi in wx[1:]]
		Cw = wx[0]*u + ec_sum([wxi*mi*u for (wxi,mi) in zip(wx[1:],public_m)])
		# create the challenge
		c = to_challenge([g, h, Aw, Cw]+Bw)
		# create responses
		rx = [(wxi - c*xi) % o for (wxi,xi) in zip(wx,x)]
		rx0_tilde = (wx0_tilde - c*x0_tilde) % o
		return (c, rx, rx0_tilde)


def verify_pi_issue(params, iparams, u, u_prime, pi_issue, biparams=None, gamma=None, ciphertext=[], public_m=[]):
	""" verify issuance proof """
	assert iparams and u and u_prime and pi_issue and len(ciphertext)+len(public_m) > 0
	(G, o, g, h) = params
	(Cx0, X) = iparams
	assert Cx0 and len(X) >= len(ciphertext)+len(public_m)

	## mixed public / private attributes
	if len(ciphertext): 
		assert biparams and gamma
		(bCx0, bX) = biparams
		assert bCx0 and len(bX) == len(X)
		(t2, t3) = u_prime
		assert t2 and t3
		(c, rx, rx0_tilde, rb, rbx, rbx0_tilde, rr) = pi_issue
		(enc_a, enc_b) = zip(*ciphertext) 
		# compute the witnesses commitments (prove knowledge of MAC's secrets)
		Aw = rx[0]*g + rx0_tilde*h + c*Cx0
		Bw = rb*g + c*u
		Cw = rbx[0]*g + rbx0_tilde*h + c*bCx0
		Dw = rb*Cx0 + c*bCx0
		# compute the witnesses commitments (prove of correctness of Xi's)
		Ew = [rxi*h + c*Xi for (rxi,Xi) in zip(rx[1:],X)]
		Fw = [rbxi*h + c*bXi for (rbxi,bXi) in zip(rbx[1:],bX)]
		Gw = [rb*Xi + c*bXi for (Xi,bXi) in zip(X,bX)]
		# compute the witnesses commitments (prove of correctness of credentials)
		t1 = [mi*g for mi in public_m]
		Hw = rr*g + ec_sum([rbxi*enc_ai for (rbxi,enc_ai) in zip(rbx[1:],enc_a)]) + c*t2
		Iw = rr*gamma + rbx[0]*g + ec_sum([rbxi*enc_bi for (rbxi,enc_bi) in zip(rbx[1:],list(enc_b)+t1)]) + c*t3
		# check the challenge
		return c == to_challenge([g, h, Aw, Bw, Cw, Dw, Hw, Iw]+Ew+Fw+Gw)

	## only public attributes
	else:
		assert len(X) >= len(public_m)
		(c, rx, rx0_tilde) = pi_issue
		# re-compute the witnesses commitments
		Aw = rx[0]*g + rx0_tilde*h + c*Cx0
		Bw = [rxi*h + c*Xi for (rxi,Xi) in zip(rx[1:],X)]
		Cw = rx[0]*u + c*u_prime + ec_sum([rxi*mi*u for (rxi,mi) in zip(rx[1:],public_m)])
		# check challenge
		return c == to_challenge([g, h, Aw, Cw]+Bw)


def make_pi_show(params, iparams, m, r, z, u_prime, sigma):
	""" make credentials showing proof """
	(G, o, g, h) = params
	(_, X) = iparams
	print(m)
	assert len(X) >= len(m)
	(u, Cm, Cu_prime) = sigma
	# create the witnesses
	wr = o.random()
	wz = [o.random() for _ in m]
	wm = [o.random() for _ in m]
	# compute the witnesses commitments
	Aw = [wmi*u + wzi*h for (wmi,wzi) in zip(wm,wz)]
	Bw = wr*g + ec_sum([wzi*Xi for (wzi,Xi) in zip(wz,X)])
	# create the challenge
	c = to_challenge([g, h, Cu_prime, Bw]+Cm+Aw)
	# create responses
	rr = (wr + c*r) % o  # note the '+' sign 
	rm = [(wmi - c*mi) % o for (wmi,mi) in zip(wm,m)]
	rz = [(wzi - c*zi) % o for (wzi,zi) in zip(wz,z)]
	return (c, rr, rm, rz)


def verify_pi_show(params, iparams, sigma, pi_show, V):
	""" verify credentials showing proof """
	assert iparams and sigma and pi_show
	(G, o, g, h) = params
	(_, X) = iparams
	(u, Cm, Cu_prime) = sigma
	assert u and len(X) >= len(Cm) > 0 and Cu_prime
	(c, rr, rm, rz) = pi_show
	# re-compute the witnesses commitments
	Aw = [rmi*u + rzi*h + c*Cmi for (rmi,rzi,Cmi) in zip(rm,rz,Cm)]
	Bw = rr*g + ec_sum([rzi*Xi for (rzi,Xi) in zip(rz,X)]) + c*V
	# verify challenge 
	return c == to_challenge([g, h, Cu_prime, Bw]+Cm+Aw)










	








""" Mac GGM scheme """
from petlib.ec import EcGroup


def setup():
	""" generate all public parameters """
	G = EcGroup()
	o = G.order()
	g = G.generator()
	h = G.hash_to_point("mac_ggm".encode("utf8"))
	return (G, o, g, h)


def keygen(params, q):
	""" mac GGM keygen """
	assert q > 0
	(G, o, g, h) = params
	sk = [o.random() for _ in range(q+1)]
	iparams = [ski*h for ski in sk[1:]]
	return (sk, iparams)


def mac(params, sk, m):
	""" compute mac GGM """
	assert len(sk) > len(m) and len(m) > 0
	(G, o, g, h) = params
	u = o.random()*g
	Hx = sk[0] + sum([sk[i+1]*m[i] for i in range(len(m))])
	u_prime = Hx*u
	sigma = (u, u_prime)
	return sigma


def verify(params, sk, m, sigma):
	""" verify mac GGM """
	assert len(sk) > len(m) and len(m) > 0
	(G, o, g, h) = params
	(u, u_prime) = sigma
	hx = sk[0] + sum([sk[i+1]*m[i] for i in range(len(m))])
	return u != G.infinite() and u_prime == hx*u



""" A setuptools based setup module. """


from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    # This is the name of your project. 
    name='amac',  # Required

    # Versions.
    version='0.0.1',  # Required

    # One-line description.
    description='Algebraic MACs and keyed-verification anonymous credentials.',  # Required

    # Longer description,
    long_description=long_description,  # Optional

    # Denotes that the long_description is in Markdown.
    long_description_content_type='text/markdown',  # Optional

    # Link to the project's main homepage.
    url='https://pypi.org/project/amac-lib',  # Optional

    # Author of the project.
    author='Alberto Sonnino',  # Optional

    # Email address corresponding to the author listed above.
    author_email='alberto.sonnino@ucl.ac.uk',  # Optional

    # Classifiers help users find your project by categorizing it.
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Security :: Cryptography',

        # License as you wish
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

        # Specify the Python versions are supported.
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
    ],

    # This field adds keywords to the project which will appear on the
    # project page.
    #
    # Note that this is a string of words separated by whitespace, not a list.
    keywords='cryptography credentials privacy',  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),  # Required

    # This field lists other packages that the project depends on to run.
    # Any package put here will be installed by pip when the project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['petlib'],  # Optional

    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    ##extras_require={  # Optional
    ##    'dev': ['check-manifest'],
    ##    'test': ['coverage'],
    ##},

    # If there are data files included in your packages that need to be
    # installed, specify them here.
    #
    # If using Python 2.6 or earlier, then these have to be included in
    # MANIFEST.in as well.
    ##package_data={  # Optional
    ##    'sample': ['package_data.dat'],
    ##},

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files
    #
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    ##data_files=[('my_data', ['data/data_file'])],  # Optional

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    #
    # For example, the following would provide a command called `sample` which
    # executes the function `main` from this package when invoked:
    ##entry_points={  # Optional
    ##    'console_scripts': [
    ##        'sample=sample:main',
    ##    ],
    ##},

    # List additional URLs that are relevant to your project as a dict.
    #
    # This field corresponds to the "Project-URL" metadata fields:
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    #
    # Examples listed include a pattern for specifying where the package tracks
    # issues, where the source is hosted, where to say thanks to the package
    # maintainers, and where to support the project financially. The key is
    # what's used to render the link text on PyPI.
    project_urls={  # Optional
        'Paper' : 'http://www0.cs.ucl.ac.uk/staff/S.Meiklejohn/files/ccs14.pdf',
        'Bug Reports': 'https://github.com/asonnino/amac',
        'Source': 'https://github.com/asonnino/amac',
    },
)